
--------------------------------------------------------------------------------
-- Made by (Jip) Willem Wijnia
-- See also: https://gitlab.com/w.b.wijnia/rainmakers
-- 
-- Please do not remove this message when modifying the map.
--------------------------------------------------------------------------------

------------------ ------------------ ------------------ ------------------ ------------------                              
-- IMPORTS --
------------------ ------------------ ------------------ ------------------ ------------------ 

local Objectives = import('/lua/ScenarioFramework.lua').Objectives
local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua')
local ScenarioFramework = import('/lua/ScenarioFramework.lua')

local GroundFunctionality = import(ScenarioInfo.path .. 'GroundAI/GroundFunctionality.lua')
local GeneralFunctionality = import(ScenarioInfo.path .. 'GeneralFunctionality.lua');
local EventFunctionality = import(ScenarioInfo.path .. 'Events/EventFunctionality.lua');

local PlatoonTinkers = import(ScenarioInfo.path .. 'Tinkers/Platoon.lua')

local Wander = import(ScenarioInfo.path .. 'Algorithms/WanderPath.lua');

local Regions = import(ScenarioInfo.path .. "GroundAI/GroundSpawnRegions.lua");

------------------ ------------------ ------------------ ------------------ ------------------                              
-- TWEAKABLE VALUES --
------------------ ------------------ ------------------ ------------------ ------------------ 

local debug = false;

-- determines when we consider a local target a threat. Equals to the amount
-- of mass a threat is worth. If it is less, then we do not consider it 
-- a threat. To prevent us from chasing scouts :).
local threatValueThreshold = 1000;

-- determines the distance we'll search from the origin node on the
-- grid. If, say, the value is 3 then we'll keep track of a three node
-- distance in all directions from the origin node.
local threatSearchDistance = 4;

-- the objective title / description.
local objectiveTitle = "Ythotha sighted";
local objectiveDescription = "An hostile Ythotha has been sighted in your operational area. Neutralize it out before it reaches the installations.";

-- the spawn offset of the platoons. Not quite interesting.
local spawnOffsetDistance = 15;

-- the speed multiplier for the experimental. If it moves too fast,
-- the guarding looks weird / fails.
local speedMultiplier = 0.85;

-- the amount of vision and the duration of that vision
-- that the players gain when the experimental spawns.
local visionRadius = 80;
local visionDuration = 22;

-- the dialogues for when the experimental spawns.
XGG_GD1_470 = { {text = '<LOC XGG_GD1_470_010>[{i HQ}]: Commander, an enemy Ythotha has been sighted. HQ out.', vid = 'XGG_HQ_GD1_04188.sfd', bank = 'XGG', cue = 'XGG_HQ_GD1_04188', faction = 'NONE'},}
local dialoguesOnSpawn = { XGG_GD1_470 };

------------------ ------------------ ------------------ ------------------ ------------------                              
-- NON-TWEAKABLE VALUES --
------------------ ------------------ ------------------ ------------------ ------------------ 

local army = ScenarioInfo.Armies.Platoon;

------------------ ------------------ ------------------ ------------------ ------------------                              
-- FUNCTIONALITY --
------------------ ------------------ ------------------ ------------------ ------------------ 

Event = { };
Event.__index = Event;

-- initialises the event.
function Event:Setup(spawnRegions, pRnd)

    -- linking
    event = { };
    setmetatable(event, Event);

    -- retrieve all event data, such as the experimental and the platoons.
    event.pRnd = pRnd;
    event.spawnRegions = spawnRegions;
    event.data = EventFunctionality.AnalyseExperimentalPlatoons("EY", army);

    return event;
end

-- eventInfo can contain:
-- eventInfo.OtherEvents        contains all the other events available in EventSetup
-- eventInfo.strPlayers         contains all the army strings of the players.
-- eventInfo.MakeObjective      determines whether an objective is created
-- eventInfo.MakeDialogue       determines whether an dialogue is played
-- eventInfo.MakeVision         determines whether vision is generated
function Event:Launch(eventInfo)

    local pathing = ScenarioInfo.Pathing;
    local hazards = ScenarioInfo.GridThreats;
    local oppertunities = ScenarioInfo.GridObjective;

    -- determine the spawn location
    local area = self.spawnRegions[self.pRnd:GetValue()];

    ------------------ ------------------ ------------------ ------------------ ------------------                              
    -- Spawn the platoons, start their AI.

    for k, platoonData in event.data.platoons do 
        local index = k;
        local location = Regions.RandomPointInRegion(area);
        local units = GroundFunctionality.SpawnGroupOfUnits(platoonData, army, location);
        local platoon = GroundFunctionality.BuildPlatoon(army, units);

        ForkThread(
            function (...) WaitSeconds (index * 10.0); PlatoonTinkers.MoveNearestHighOppertunity(unpack(arg)); end,
            oppertunities,
            platoon
        );
    end

    ------------------ ------------------ ------------------ ------------------ ------------------                              
    -- Spawn the experimental and start its AI. Slow it down slightly such that the platoons can keep up.

    local location = Regions.RandomPointInRegion(area);
    local experimental = GroundFunctionality.SpawnUnit(self.data.experimental, army, location);
    experimental:SetSpeedMult(speedMultiplier);

    function ExperimentalBehaviour(experimental, oppertunities)

        local target = nil;

        while not experimental:IsDead() do

            local position = experimental:GetPosition();

            -- check for local oppertunities
            local range = 50;
            local value = 2500;
            local ops = hazards:FindLocalOppertunities(position, range, value);
            local count = table.getn(ops);

            if count > 0 then 
                -- move the experimental away from the targets. Keep track of the fact that we're falling back.
                IssueClearCommands({experimental});
                local locationToEngage = PlatoonTinkers.ComputeEngageLocation(position, ops);

                local totalHealth = experimental:GetMaxHealth();
                local currentHealth = experimental:GetHealth();

                -- are we mad or not mad? That is the question
                if currentHealth / totalHealth > 0.5 then
                    IssueAggressiveMove({experimental}, locationToEngage)
                else
                    IssueMove({experimental}, locationToEngage)
                end
            end

            if experimental:IsIdleState() then 

                ------------------ ------------------ ------------------ ------------------ ------------------   
                -- check if we have anywhere to go
                
                if not oppertunities:OppertunitiesAvailable() then 
                    if oppertunities == ScenarioInfo.GridObjective then 
                        LOG("Experimental cannot find new oppertunities - switching to player oppertunities.");
                        oppertunities = ScenarioInfo.GridPlayers;

                    else
                        LOG("Experimental cannot find new oppertunities.");
                        break;
                    end
                end

                -- find a target 
                target = oppertunities:GetNearestHighOppertunity(position);

                if target then 

                    -- find a path
                    local options = { smoothen = true, force = 15000 };
                    local path = pathing:FindPath(position, target.position, options)

                    -- move in an aggressive, angry looking fashion
                    IssueClearCommands({experimental});
                    local count = table.getn(path);
                    for k = 1, count - 2, 1 do 
                        IssueMove({experimental}, path[k]);
                    end

                    local totalHealth = experimental:GetMaxHealth();
                    local currentHealth = experimental:GetHealth();

                    -- are we mad or not mad? That is the question
                    if currentHealth / totalHealth > 0.5 then
                        IssueAggressiveMove({experimental}, target.position)
                    else
                        IssueMove({experimental}, target.position)
                    end
                end
            else
                -- check if we should reconsider
                if target then 
                    local node = oppertunities:GetNode(target.position);
                    if node.oppertunity < 0.5 * target.oppertunity  then
                        -- if the target isn't what it was a few seconds ago, make the platoon stop
                        LOG("An experimental is reconsidering.");
                        IssueClearCommands({experimental});
                    end
                end
            end

            WaitSeconds(1.0);
        end
    end

    ForkThread(ExperimentalBehaviour, experimental, oppertunities);

    ------------------ ------------------ ------------------ ------------------ ------------------                              
    -- Give it a real campaign feeling to it.

    -- sound notification
    if eventInfo.MakeDialogue then
        EventFunctionality.PlayRandomDialogue(dialoguesOnSpawn);
    end

    -- objective notification
    if eventInfo.MakeObjective then
        self.ObjectiveColossus = Objectives.Kill(
            'secondary', 
            'incomplete', 
            objectiveTitle,        
            objectiveDescription, 
            {                          
                Units = { experimental },
                MarkUnits = true,
            }
        )
    end

    -- give temporarily vision.
    if eventInfo.MakeVision then
        for k, player in eventInfo.StrPlayers do
            ScenarioFramework.CreateVisibleAreaAtUnit(visionRadius, experimental, visionDuration, GetArmyBrain(player));
        end
    end
end