
local AIAttackUtils = import('/lua/ai/aiattackutilities.lua');

local HeapAsset = import(ScenarioInfo.path .. "Algorithms/Heap.lua");
local BezierAsset = import(ScenarioInfo.path .. "Algorithms/BezierSpline.lua");

-- we want to roughly know what a cell contains:
-- a commander
-- mobile units, and their tech
-- static defenses, and their tech
-- other structures

-- and we do not count:
-- walls

Pathing = { };
Pathing.__index = Pathing;

------------------ ------------------ ------------------ ------------------ ------------------                              
-- constructs a node with typical properties

function ConstructNode(x, z)
    return { 
        nearestPathingNode = nil,
        namePathingNode = '',
        namePathingGraph = ''
    }
end

function Pathing:Setup(grid, layer)

    local pathing = { };
    setmetatable(pathing, Pathing);

    ------------------ ------------------ ------------------ ------------------ ------------------                              
    -- store all the parameters

    pathing.grid = grid;
    pathing.layer = layer;

    -- save the pathing nodes
    pathing.graphs = AIAttackUtils.GetPathGraphs()[layer];
    if not pathing.graphs then
        error("No land nodes found.");
    end

    -- find the nearest (land) pathing layer for each node in the grid
    for z = 0, grid.precision - 1 do  
        for x = 0, grid.precision - 1 do

            local index         = grid:GetIndex(x, z);
            local node          = grid.nodes[index];
            local position      = grid:NodeToCoordinates(node);

            -- find the nearest marker
            local marker        = nil;
            local markerName    = '';
            local graphName     = '';
            local distance      = 10000000.0;

            for gn, graph in pathing.graphs do
                for mn, markerInfo in graph do
                    local dist2 = VDist2Sq(position[1], position[3], markerInfo.position[1], markerInfo.position[3])
                    if dist2 < distance then 
                        distance = dist2;

                        marker = markerInfo;
                        markerName = mn;
                        graphName = gn;
                    end
                end
            end

            node.nearestPathingNode = marker;
            node.namePathingNode = markerName;
            node.namePathingGraph = graphName;
        end
    end

    -- add the marker name to all markers
    for name, graph in pathing.graphs do
        for mn, node in graph do
            node.name = mn;
        end
    end

    -- pre-compute distances for adjacent nodes
    for name, graph in pathing.graphs do
        for mn, node in graph do 

            local neighbours = { };
            for a, neighbour in node.adjacent do 
                local data = { };
                data.neighbour = neighbour;
                data.distance = VDist2(
                    node.position[1], 
                    node.position[3], 
                    graph[neighbour].position[1], 
                    graph[neighbour].position[3]
                );
                table.insert(neighbours, data);
            end
    
            node.neighbours = neighbours;
        end
    end

    -- add data to each pathing node
    -- (these are _not_ the nodes generated by the grid)
    for name, graph in pathing.graphs do
        for mn, node in graph do
            node.neighbouringHazard = 0; 
            node.value = 0;
            node.hazard = 0;
            node.visited = false;
            node.predictedCosts = 0;
            node.pastCosts = 0;
            node.parent = nil;
        end
    end

    return pathing;

end

------------------ ------------------ ------------------ ------------------ ------------------                              
-- given a point and a grid, retrieve the closest pathing node

function Pathing:GetPathingNode(point)

    -- make coordinates relative to the grid
    local gx = point[1] - self.grid.areaLeftTop[1];
    local gz = point[3] - self.grid.areaLeftTop[3];

    -- turn them into (integral) grid coordinates
    local ix = math.floor(gx / self.grid.stepSizeX + 0.5);
    local iz = math.floor(gz / self.grid.stepSizeZ + 0.5)

    -- retrieve the vertex
    local index = self.grid:GetIndex(ix, iz);
    local vertex = self.grid.nodes[index];

    -- check if we actually have one
    if not vertex then 
        error('Cannot find a vertex for point: ' .. repr(point));
    end

    -- return the node
    return vertex.nearestPathingNode, vertex;
end

------------------ ------------------ ------------------ ------------------ ------------------                              
-- given a pathing node, retrieves the neighbouring pathing nodes

function Pathing:GetPathingNodeNeighbours(graph, node) 
    local neighbours = { };
    for k, adjancendee in node.neighbours do 
        local neighbour = graph[adjancendee.neighbour];
        table.insert(neighbours, neighbour);
    end

    return neighbours;
end

------------------ ------------------ ------------------ ------------------ ------------------                              
-- keeps track of / updates / manages a grid

function Pathing:Tick()

    LOG("Starting a pathing tick for layer \'" .. repr(self.layer) .. "\'.");

    -- lessens the hazard level on all pathing nodes
    function UpdateNodes(graphs)
        for name, graph in graphs do
            for mn, node in graph do 
                node.hazard = node.hazard * 0.1;
                node.neighbouringHazard = node.neighbouringHazard * 0.1;
            end
        end
    end

    -- increases the hazard level on all pathing nodes
    function UpdateHazard(graphs, vertices)
        for k, vertex in vertices do 

            -- update local hazard
            local node = vertex.nearestPathingNode;
            node.hazard = node.hazard + vertex.hazard;  

            -- update neighbouring hazard
            local graph = vertex.namePathingGraph;
            for l, data in node.neighbours do 
                local neighbour = graphs[graph][data.neighbour];
                neighbour.neighbouringHazard = neighbour.neighbouringHazard + 0.1 * vertex.hazard / data.distance;
            end
        end
    end

    -- attempt to make sure that ticks do not happen on the same sim step
    WaitSeconds(Random() * 5.0);

    while true do
        UpdateNodes(self.graphs);
        UpdateHazard(self.graphs, self.grid.nodes);
        WaitSeconds(4.0);
    end
end

-- Finds a path from the origin towards the destination.
-- Options = {
    -- smoothen = true / false,                 -- whether or not the path will be smoothen through applying a bezier curve
    -- hazardMultiplier = float,                -- what to do with hazards on the path: avoide them (positive) ignore them (all hazards = 0), or move through hazards (negative)
    -- force = float                            -- the (military) strength of the unit, the larger it is the less hazard has an influence
-- }
function Pathing:FindPath(origin, destination, options)

    -- sane defaults
    local value = 10.0;
    local smoothen = true;
    local hazardMultiplier = 1.0;

    -- interpret the options set
    if options then
        value = options.force or value;
        smoothen = options.smoothen or smoothen;
        hazardMultiplier = options.hazardMultiplier or hazardMultiplier;
    end

    -- keep track of the nodes that we've visited
    local visited = { }

    -- checks if two nides are equal
    function Equal(a, b)
        return a.name == b.name;
    end

    -- get the origin and desination nodes
    local pon, on = self:GetPathingNode(origin);
    local pdn, dn = self:GetPathingNode(destination);

    if not pon then
        error("Pathfinding: origin is not part of the grid")
        return { }
    end

    if not pdn then
        error("Pathfinding: destination is not part of the grid")
        return { }
    end 

    if Equal(pon, pdn) then
        return { };
    end

    if not on.namePathingGraph == dn.namePathingGraph then 
        error("Pathfinding: nodes are not part of the same graph");   
    end

    local graph = self.graphs[on.namePathingGraph];
    
    -- construct a heap: it will keep things O(nlgn).
    local compare = function(node) return node.pastCosts + node.predictedCosts end 
    local heap = HeapAsset.MinHeap:Initiate(compare);

    -- start the search
    pon.visited = true;
    pon.value = value;
    heap:Insert(pon);
    table.insert(visited, pon);

    while not heap:Empty() do 

        -- -- visualize it
        -- WaitSeconds(0.1);

        -- for k, visitee in visited do 
        --     DrawCircle(visitee.position, 10, "0000ff");
        --     DrawCircle(visitee.position, math.max(1, math.log(visitee.value)) + 0.15, "00ff00");
        -- end

        -- retrieve our current node
        local node = heap:ExtractMin();

        -- did we find the node we are looking for?
        if Equal(node, pdn) then 
            break;
        end

        -- check its neighbours
        local neighbourData = node.neighbours;
        for k, data in neighbourData do 
            local neighbour = graph[data.neighbour];
            -- do not visit nodes twice
            if not neighbour.visited then 

                -- determine the (remaining) value / hazard when we enter this node
                local hazard = node.hazard + node.neighbouringHazard;
                neighbour.value = math.max(0, node.value - hazard);     -- do not let it be negative
                hazard = math.max(0, hazard - node.value);              -- do not let it be negative

                neighbour.visited = true;
                neighbour.parent = node;
                neighbour.predictedCosts = VDist2(neighbour.position[1], neighbour.position[3], pdn.position[1], pdn.position[3]);
                neighbour.pastCosts = 
                    node.pastCosts + 
                    data.distance +
                    hazard;

                -- store it in the heap
                heap:Insert(neighbour);

                -- keep track of the nodes we've changed
                table.insert(visited, neighbour);
            end
        end
    end

    -- did we find a path?
    if not pdn.parent then 
        return { }
    end

    -- retrieve the (reversed) path
    local revPath = { };
    local node = pdn;
    while node.parent do 
        table.insert(revPath, node.position);

        -- make duplicates to make it follow the original path more
        node = node.parent;
    end

    -- reverse the path
    local path = { };
    for k = table.getn(revPath), 1, -1 do 
        table.insert(path, revPath[k]);
    end

    -- clean up after ourselves
    for k, node in visited do 
        node.visited = false;
        node.value = 0;
        node.parent = nil;
        node.predictedCosts = 0;
        node.pastCosts = 0;
    end

    -- smooth the path
    if smoothen then 

        -- double it, such that it follows the original path more
        pathDouble = { };
        for k, node in path do 
            table.insert(pathDouble, node);
            table.insert(pathDouble, node);
        end

        -- smoothen it
        pathDouble = BezierAsset.SmoothPath(pathDouble);

        -- and remove the 'duplicates'
        path = { };
        for k, node in pathDouble do 
            if math.floor(k * 0.5) * 2 == k then 
                table.insert(path, node);
            end
        end
    end

    return path;
end

------------------ ------------------ ------------------ ------------------ ------------------                              
-- hurrrr, debug

function Pathing:DrawPathingNodes()
    while true do 
        WaitSeconds(0.1);
        for name, graph in self.graphs do
            for mn, node in graph do
                DrawCircle( node.position, 1, "ffffff" );
            end
        end
    end
end

function Pathing:DrawPathingNodesHazard()
    while true do 
        WaitSeconds(0.1);
        for name, graph in self.graphs do
            for mn, node in graph do
                DrawCircle( node.position, math.max(1, math.log(node.hazard + node.neighbouringHazard)) + 0.15, "ff0000" );
            end
        end
    end
end

function Pathing:DrawNearestMarker()
    while true do 
        WaitSeconds(0.1);
        for k, vertex in self.grid.nodes do 
            local vertexPosition = self.grid:NodeToCoordinates(vertex);
            local node = vertex.nearestPathingNode;
            local nodePosition = node.position;
            DrawLinePop( vertexPosition, nodePosition, "cccccc" );
        end
    end
end

function Pathing:DrawPathingConnections()
    while true do 
        WaitSeconds(0.1);
        for name, graph in self.graphs do
            for mn, node in graph do
                local neighbours = self:GetPathingNodeNeighbours(graph, node);
                for k, neighbour in neighbours do 
                    DrawLinePop( node.position, neighbour.position, "cccccc" );
                end
            end
        end
    end
end